"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var uml = require("./uml/index");
var Delinter = (function () {
    function Delinter() {
        this._umlCodeModel = new uml.CodeModel();
    }
    Object.defineProperty(Delinter.prototype, "umlCodeModel", {
        /**
         * Uml code model description filled by the parse method(s)
         *
         * @readonly
         * @type {uml.CodeModel}
         * @memberOf Delinter
         */
        get: function () {
            return this._umlCodeModel;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Delint a TypeScript source file, adding the parsed elements to umlCodeModel.
     *
     * @param {ts.SourceFile} file TypeScript source file
     *
     * @memberOf Delinter
     */
    Delinter.prototype.parse = function (file) {
        this._delintNode(file);
    };
    Delinter.prototype._delintNode = function (node) {
        var _this = this;
        switch (node.kind) {
            case ts.SyntaxKind.ClassDeclaration:
                this._delintClass(node, this._delintClassNode.bind(this));
                break;
            case ts.SyntaxKind.InterfaceDeclaration:
                this._delintClass(node, this._delintInterfaceNode.bind(this));
                break;
            default:
                ts.forEachChild(node, function (n) { _this._delintNode(n); });
                break;
        }
    };
    Delinter.prototype._delintClass = function (node, delintChild) {
        var stereotype = 0 /* None */;
        if (node.kind === ts.SyntaxKind.InterfaceDeclaration) {
            stereotype = 2 /* Interface */;
        }
        var umlClass = new uml.Class(node.name.getText(), stereotype);
        this._umlCodeModel.nodes.setValue(umlClass.identifier, umlClass);
        this._delintHeritageClauses(node.heritageClauses, umlClass);
        ts.forEachChild(node, function (n) { delintChild(n, umlClass); });
    };
    Delinter.prototype._delintClassNode = function (node, umlClass) {
        var _this = this;
        switch (node.kind) {
            case ts.SyntaxKind.PropertyDeclaration:
                this._delintProperty(node, umlClass);
                break;
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.SetAccessor:
                this._delintClassGetterSetter(node, umlClass);
                break;
            case ts.SyntaxKind.MethodDeclaration:
                this._delintMethod(node, umlClass);
                break;
            default:
                ts.forEachChild(node, function (n) { _this._delintClassNode(n, umlClass); });
                break;
        }
    };
    Delinter.prototype._delintInterfaceNode = function (node, umlClass) {
        var _this = this;
        switch (node.kind) {
            case ts.SyntaxKind.PropertySignature:
                this._delintProperty(node, umlClass);
                break;
            case ts.SyntaxKind.MethodSignature:
                this._delintMethod(node, umlClass);
                break;
            default:
                ts.forEachChild(node, function (n) { _this._delintClassNode(n, umlClass); });
                break;
        }
    };
    Delinter.prototype._delintHeritageClauses = function (heritageClauses, umlClass) {
        var _this = this;
        if (heritageClauses) {
            heritageClauses.forEach(function (h) {
                switch (h.token) {
                    case ts.SyntaxKind.ImplementsKeyword:
                        h.types.forEach(function (t) {
                            var interfaceName = t.expression.getText();
                            // Add interface to CodeModel if not exists
                            if (!_this._umlCodeModel.nodes.containsKey(interfaceName)) {
                                var umlInterface = new uml.Class(interfaceName, 2 /* Interface */);
                                _this._umlCodeModel.nodes.setValue(interfaceName, umlInterface);
                            }
                            var generalization = new uml.Generalization(umlClass.identifier, interfaceName);
                            _this._umlCodeModel.generalizations.add(generalization);
                        });
                        break;
                    case ts.SyntaxKind.ExtendsKeyword:
                        h.types.forEach(function (t) {
                            var parentClassName = t.expression.getText();
                            // Add interface to CodeModel if not exists
                            if (!_this._umlCodeModel.nodes.containsKey(parentClassName)) {
                                var umlParentClass = new uml.Class(parentClassName);
                                _this._umlCodeModel.nodes.setValue(parentClassName, umlParentClass);
                            }
                            var generalization = new uml.Generalization(umlClass.identifier, parentClassName);
                            _this._umlCodeModel.generalizations.add(generalization);
                        });
                        break;
                    /* istanbul ignore next: default case never reached */
                    default:
                        break;
                }
            });
        }
    };
    Delinter.prototype._delintMethod = function (methodDeclaration, umlClass) {
        var _this = this;
        var identifier = methodDeclaration.name.getText();
        // Default to public accessibility
        var accessibility = this._delintAccessibilityModifiers(methodDeclaration.modifiers);
        var method = new uml.FunctionProperty(identifier, accessibility);
        if (methodDeclaration.type) {
            method.returnType = this._delintType(methodDeclaration.type);
        }
        methodDeclaration.parameters.forEach(function (p) {
            var parameter = new uml.Parameter(p.name.getText());
            parameter.type = _this._delintType(p.type);
            if (p.initializer) {
                parameter.defaultInitializer = p.initializer.getText();
                parameter.optional = true;
            }
            if (p.questionToken) {
                parameter.optional = true;
            }
            method.parameters.push(parameter);
        });
        method.optional = methodDeclaration.questionToken !== undefined;
        umlClass.methods.setValue(identifier, method);
    };
    Delinter.prototype._delintClassGetterSetter = function (node, umlClass) {
        var _this = this;
        var identifier = node.name.getText();
        // Default to public accessibility
        var accessibility = this._delintAccessibilityModifiers(node.modifiers);
        var type;
        if (node.kind === ts.SyntaxKind.GetAccessor) {
            type = this._delintType(node.type);
        }
        else {
            if (node.parameters.length > 0) {
                type = this._delintType(node.parameters[0].type);
            }
        }
        var variable = new uml.VariableProperty(identifier, accessibility, type);
        // Set stereotype based on kind
        var existingVariable = umlClass.variables.getValue(identifier);
        var stereotype;
        if (node.kind === ts.SyntaxKind.GetAccessor) {
            // If a setter already exists, use stereotype GetSet
            if (existingVariable && existingVariable.stereotype === 4 /* Set */) {
                stereotype = 5 /* GetSet */;
            }
            else {
                stereotype = 3 /* Get */;
            }
        }
        else {
            // If a setter already exists, use stereotype GetSet
            if (existingVariable && existingVariable.stereotype === 3 /* Get */) {
                stereotype = 5 /* GetSet */;
            }
            else {
                stereotype = 4 /* Set */;
            }
        }
        variable.stereotype = stereotype;
        umlClass.variables.setValue(identifier, variable);
        this._typeAssociations(umlClass.identifier, type).forEach(function (a) {
            _this.umlCodeModel.associations.add(a);
        });
    };
    Delinter.prototype._delintProperty = function (property, umlInterface) {
        var _this = this;
        var identifier = property.name.getText();
        // Default to public accessibility
        var accessibility = this._delintAccessibilityModifiers(property.modifiers);
        var type = this._delintType(property.type);
        var variable = new uml.VariableProperty(identifier, accessibility, type);
        variable.optional = property.questionToken !== undefined;
        umlInterface.variables.setValue(identifier, variable);
        this._typeAssociations(umlInterface.identifier, type).forEach(function (a) {
            _this.umlCodeModel.associations.add(a);
        });
    };
    Delinter.prototype._delintAccessibilityModifiers = function (modifiers) {
        var accessibility = 0 /* Public */;
        if (modifiers) {
            modifiers.forEach(function (m) {
                switch (m.kind) {
                    case ts.SyntaxKind.PrivateKeyword:
                        accessibility = 2 /* Private */;
                        break;
                    case ts.SyntaxKind.ProtectedKeyword:
                        accessibility = 1 /* Protected */;
                        break;
                    case ts.SyntaxKind.PublicKeyword:
                        accessibility = 0 /* Public */;
                        break;
                }
            });
        }
        return accessibility;
    };
    Delinter.prototype._delintType = function (typeNode) {
        var type = null;
        var kind = typeNode ? typeNode.kind : null;
        switch (kind) {
            case ts.SyntaxKind.AnyKeyword:
            case ts.SyntaxKind.NumberKeyword:
            case ts.SyntaxKind.BooleanKeyword:
            case ts.SyntaxKind.StringKeyword:
            case ts.SyntaxKind.SymbolKeyword:
            case ts.SyntaxKind.VoidKeyword:
                type = new uml.PrimaryType(typeNode.getText(), 1 /* PredefinedType */);
                break;
            case ts.SyntaxKind.TypeReference:
                var typeReference = typeNode;
                type = new uml.PrimaryType(typeReference.getText(), 2 /* TypeReference */);
                type.name = typeReference.typeName.getText();
                var typeRef = typeNode;
                type.typeArguments = this._delintTypeArguments(typeRef.typeArguments);
                break;
            case ts.SyntaxKind.TypeLiteral:
                type = new uml.PrimaryType("TypeLiteral", 3 /* ObjectType */);
                break;
            case ts.SyntaxKind.TupleType:
                type = new uml.PrimaryType(typeNode.getText(), 5 /* TupleType */);
                var tupleType = typeNode;
                type.typeArguments = this._delintTypeArguments(tupleType.elementTypes);
                break;
            case ts.SyntaxKind.ArrayType:
                type = new uml.PrimaryType(typeNode.getText(), 4 /* ArrayType */);
                break;
            case ts.SyntaxKind.TypeQuery:
                type = new uml.PrimaryType(typeNode.getText(), 6 /* TypeQuery */);
                break;
            case ts.SyntaxKind.ThisType:
                type = new uml.PrimaryType(typeNode.getText(), 7 /* ThisType */);
                break;
            case ts.SyntaxKind.UnionType:
                type = new uml.UnionOrIntersectionType(typeNode.getText(), 100 /* Union */);
                var union = typeNode;
                type.types = this._delintTypeArguments(union.types);
                break;
            case ts.SyntaxKind.IntersectionType:
                type = new uml.UnionOrIntersectionType(typeNode.getText(), 101 /* Intersection */);
                var intersection = typeNode;
                type.types = this._delintTypeArguments(intersection.types);
                break;
            default:
                type = new uml.PrimaryType("any", 0 /* ImplicitAny */);
                break;
        }
        return type;
    };
    Delinter.prototype._delintTypeArguments = function (typeArguments) {
        var _this = this;
        var delintedTypes = [];
        if (typeArguments) {
            delintedTypes = typeArguments.map(function (value) {
                return _this._delintType(value);
            });
        }
        return delintedTypes;
    };
    Delinter.prototype._typeAssociations = function (from, type) {
        var _this = this;
        // Add association to type
        var associations = [];
        if (type) {
            if (type instanceof uml.PrimaryType || type instanceof uml.UnionOrIntersectionType) {
                switch (type.kind) {
                    case 2 /* TypeReference */:
                        associations = [new uml.Association(from, type.name)];
                    // Fall through to process potential (generic) type arguments
                    case 4 /* ArrayType */:
                    case 5 /* TupleType */:
                        type.typeArguments.forEach(function (t) {
                            associations = associations.concat(_this._typeAssociations(from, t));
                        });
                        break;
                    case 100 /* Union */:
                    case 101 /* Intersection */:
                        type.types.forEach(function (t) {
                            associations = associations.concat(_this._typeAssociations(from, t));
                        });
                        break;
                    default:
                        // Do not produce associations
                        break;
                }
            }
        }
        return associations;
    };
    return Delinter;
}());
exports.Delinter = Delinter;
